﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq.Expressions;
using System.Runtime.CompilerServices;
using System.Threading;
using System.Threading.Tasks;
using System.Timers;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using CH.Regatron.HPPS.Device;
using CH.Regatron.HPPS.SolarArraySimulation;
using CH.Regatron.HPPS.TFE;
using TopConAPIDotNet4WPFExample.Commands;
using Timer = System.Timers.Timer;

namespace TopConAPIDotNet4WPFExample.SasSimulation
{
  public class SasSimulationViewModel : INotifyPropertyChanged
  {
    #region //---- Members ----

    private ComboBox _curveEditorComboBox;
    private readonly FunctionEngine _functionEngine;
    private readonly CH.Regatron.HPPS.Device.TopCon _topConDev;

    private TfeCurveItem _actualTfeCurveItem;

    private ICommand _loadListCommand;
    private ICommand _stopLiveViewerCommand;
    private ICommand _runLiveViewerCommand;

    private double _irradianceReference;
    private double _irradianceDuration;
    private double _mppVoltage;
    private double _mppCurrent;
    private double _mppPower;
    private double _irradianceActual;
    private double _temperatureActual;
    private double _fadetimeRampeTime;
    private string _message;

    private CancellationTokenSource _tokenSource;
    CancellationToken _cancelToken;
    private const int UpdateIntervallMs = 1000;
    private Task _valueUpdateTask;

    #endregion

    #region //---- Properties ----

    public ComboBox CurveEditorComboBox
    {
      get => _curveEditorComboBox;
      set => _curveEditorComboBox = value;
    }

    public bool IsConnected
    {
      get
      {
        if (_topConDev != null)
        {
          return _topConDev.IsConnected();
        }
        else
        {
          return false;
        }
      }
    }

    public bool IsSasSimAvailable
    {
      get
      {
        if (_topConDev != null)
        {
          if (_topConDev.IsConnected())
          {
            return _topConDev.IsOptionSasEnabled() && _topConDev.IsOptionTfeEnabled();
          }
          else
          {
            return false;
          }
        }
        else
        {
          return false;
        }
      }
    }

    private bool TopIsInReadyState { get; set; }

    private bool Running { get; set; }

    public double IrradianceReference
    {
      get
      {
        return _irradianceReference;
      }

      set
      {
        _irradianceReference = value;
        OnPropertyChanged(nameof(IrradianceReference));
      }
    }

    public double IrradianceDuration
    {
      get
      {
        return _irradianceDuration;
      }

      set
      {
        _irradianceDuration = value;
        OnPropertyChanged(nameof(IrradianceDuration));
      }
    }

    public double MppVoltage
    {
      get
      {
        return _mppVoltage;
      }

      set
      {
        _mppVoltage = value;
        OnPropertyChanged(nameof(MppVoltage));
      }
    }

    public double MppCurrent
    {
      get
      {
        return _mppCurrent;
      }

      set
      {
        _mppCurrent = value;
        OnPropertyChanged(nameof(MppCurrent));
      }
    }

    public double MppPower
    {
      get
      {
        return _mppPower;
      }

      set
      {
        _mppPower = value;
        OnPropertyChanged(nameof(MppPower));
      }
    }

    public double IrradianceActual
    {
      get
      {
        return _irradianceActual;
      }

      set
      {
        _irradianceActual = value;
        OnPropertyChanged(nameof(IrradianceActual));
      }
    }

    public double TemperatureActual
    {
      get
      {
        return _temperatureActual;
      }

      set
      {
        _temperatureActual = value;
        OnPropertyChanged(nameof(TemperatureActual));
      }
    }

    public double FadetimeRampeTime
    {
      get
      {
        return _fadetimeRampeTime;
      }

      set
      {
        _fadetimeRampeTime = value;
        OnPropertyChanged(nameof(FadetimeRampeTime));
      }
    }

    public string Message
    {
      get
      {
        return _message;
      }

      set
      {
        _message = value;
        OnPropertyChanged(nameof(Message));
      }
    }

    private bool CurveInitialized { get; set; }

    private uint CurrentSequenceNumber { get; set; }

    #endregion

    public SasSimulationViewModel(CH.Regatron.HPPS.Device.TopCon topCon)
    {
      _functionEngine = topCon.GetFunctionEngine();

      topCon.SetPowerOff();
      topCon.WatchdogRs232Clear();
      topCon.ClearError();
      _topConDev = topCon;
      TopIsInReadyState = topCon.IsInReadyState();
      IrradianceReference = 100;
      IrradianceDuration = 60;
      MppVoltage = 0;
      MppCurrent = 0;
      MppPower = 0;
      IrradianceActual = 0;
      TemperatureActual = 0;
      FadetimeRampeTime = 0;

      Message = string.Empty;

      if (_functionEngine != null)
      {
        Message = "Click on Load Flash List.";
      }
      else
      {
        Message = "Functions generator option not installed in TC device";
      }

      UdateConnectionStateProperties();
    }

    public void UdateConnectionStateProperties()
    {
      OnPropertyChanged(nameof(IsConnected));
      OnPropertyChanged(nameof(IsSasSimAvailable));
    }

    #region //---- Value Updater ----

    private void ValueUpdateWorker(CancellationToken cancelToken)
    {
      try
      {
        while(!_cancelToken.IsCancellationRequested)
        {
          if (IsConnected)
          {
            if(CurveInitialized == false)
            {
              _functionEngine.SwitchToCurve(CurrentSequenceNumber);
              Message = $"Switching to curve {CurrentSequenceNumber}";
              Thread.Sleep(1000);
              //_topCon.SetPowerOn();  //-- switch power on

              //Message = "Power is on";

              //-- if not using the new curve, switch to that curve
              _functionEngine.SwitchToCurve(CurrentSequenceNumber);
              Message = $"Switching to curve {CurrentSequenceNumber}";
              Thread.Sleep(1000);
              //-- Switch to a new irradiance level within a longer time  (==>  we can monitor the transition)
              _functionEngine.SetSasIrradianceReference(IrradianceReference, IrradianceDuration);
              Message = string.Format("Changing irradiance to {0} W/m2 within {1} s", IrradianceReference, IrradianceDuration);

              CurveInitialized = true;
            }

            MPPData mppData = _functionEngine.GetSasMppData();
            MppVoltage = Math.Round(mppData.GetMppVoltage(), 4);
            MppCurrent = Math.Round(mppData.GetMppCurrent(), 4);
            MppPower = Math.Round(mppData.GetMppPower(), 4);

            IrradianceActual = Math.Round(_functionEngine.GetSasIrradianceActual(), 4);
            TemperatureActual = Math.Round(_functionEngine.GetSasTemperatureActual(), 4);
            FadetimeRampeTime = Math.Round(_functionEngine.GetSasTemperatureRampTimeRemaining(), 4);
          }

          Task.Delay(UpdateIntervallMs, _cancelToken).Wait(_cancelToken);
        }

        Debug.WriteLine("ValueUpdater ended");
      }
      catch(Exception e)
      {
        Console.WriteLine(e);
      }
    }

    private bool IsValueUpdateTaskActiv()
    {
      return _valueUpdateTask?.Status == TaskStatus.Running;
    }

    private void StartValueUpdateTask()
    {
      if(_tokenSource == null)
      {
        _tokenSource = new CancellationTokenSource();
        _cancelToken = _tokenSource.Token;
      }

      if(IsConnected && !IsValueUpdateTaskActiv())
      {
        _valueUpdateTask = Task.Factory.StartNew(() => ValueUpdateWorker(_cancelToken), _cancelToken, TaskCreationOptions.LongRunning, TaskScheduler.Default);
      }
    }

    private void StopValueUpdateTask()
    {
      if(IsValueUpdateTaskActiv())
      {
        _tokenSource.Cancel();

        try
        {
          _valueUpdateTask.Wait();

          _tokenSource = null;
          _valueUpdateTask = null;
        }
        catch(Exception e)
        {
          if(e?.InnerException is TaskCanceledException || e is OperationCanceledException)
          {
            //dont do anything
          }
          else
          {
            throw;
          }
        }
      }
    }

    #endregion

    #region //---- Callbacks ----

    private void CurveEditorComboBox_SelectionChanged(object sender, SelectionChangedEventArgs e)
    {
      StopValueUpdateTask();
      var comboBox = sender as ComboBox;
      if(comboBox != null)
      {
        _actualTfeCurveItem = comboBox.SelectedValue as TfeCurveItem;
      }
    }

    #endregion
    
    #region //---- Commands ----

    #region //-- LoadListCommand --

    public ICommand LoadListCommand
    {
      get
      {
        if(_loadListCommand == null && _functionEngine != null)
        {
          _loadListCommand = new GenericCommand(DoLoadList);
        }
        return _loadListCommand;
      }
    }

    private void DoLoadList(object obj)
    {
      try
      {
        Mouse.OverrideCursor = Cursors.Wait;

        StopValueUpdateTask();
        if(IsConnected && _functionEngine != null)
        {
          Message = string.Empty;
          //Clear combobox
          CurveEditorComboBox.Items.Clear();

          try
          {
            _functionEngine.LoadCurveListFromTopCon();
          }
          catch (TopConException topConException)
          {
            MessageBox.Show($"Error in Load Curve List: \n{topConException.Message}", "Load Curve List", MessageBoxButton.OK, MessageBoxImage.Error);
            return;
          }
          int itemsCount = 0;
          _functionEngine.ForEach(delegate (TFECurve curve)
          {
            itemsCount++;
            TfeCurveItem tfeCurveItem = new TfeCurveItem(curve);
            CurveEditorComboBox.Items.Add(tfeCurveItem);
          });

          if(itemsCount > 0)
          {
            CurveEditorComboBox.SelectionChanged += CurveEditorComboBox_SelectionChanged;
            CurveEditorComboBox.SelectedIndex = 0;
            Message = string.Empty;
          }
          else
          {
            Message = "No curves found, please use SASControl and add curves!!!!!";
          }
        }
      }
      finally
      {
        Mouse.OverrideCursor = null;
      }
      
    }

    #endregion

    #region //-- RunLiveviewerCommand --

    public ICommand RunLiveViewerCommand
    {
      get
      {
        if(_runLiveViewerCommand == null)
        {
          _runLiveViewerCommand = new GenericCommand(RunLiveViewer, CanRunLiveViewer);
        }
        return _runLiveViewerCommand;
      }
    }

    private void RunLiveViewer(object obj)
    {
      if(IsConnected)
      {
        if(_actualTfeCurveItem != null)
        {
          CurveInitialized = false;
          CurrentSequenceNumber = _actualTfeCurveItem.SeqNo;
          StartValueUpdateTask();
          Running = true;
        }
      }
    }

    private bool CanRunLiveViewer(object obj)
    {
      if(IsConnected && _actualTfeCurveItem != null && Running == false)
      {
        return true;
      }
      return false;
    }

    #endregion

    #region //-- StopLiveviewerCommand --

    public ICommand StopLiveViewerCommand
    {
      get
      {
        if(_stopLiveViewerCommand == null)
        {
          _stopLiveViewerCommand = new GenericCommand(StopLiveViewer, CanStopLiveViewer);
        }
        return _stopLiveViewerCommand;
      }
    }

    private void StopLiveViewer(object obj)
    {
      StopLiveViewer();
    }

    public void StopLiveViewer()
    {
      Running = false;
      StopValueUpdateTask();
    }

    private bool CanStopLiveViewer(object obj)
    {
      if(IsConnected && Running == true)
      {
        return true;
      }
      return false;
    }

    #endregion

    #endregion

    #region //---- IPropertyChanged Interface ----

    public event PropertyChangedEventHandler PropertyChanged;

    private void OnPropertyChanged(string propertyName)
    {
      PropertyChangedEventHandler handler = PropertyChanged;
      if(handler != null) handler(this, new PropertyChangedEventArgs(propertyName));
    }

    #endregion
  }
}
